from __future__ import print_function, absolute_import

from . import HGui, Connection, Node, Hook, Attribute
import weakref
from bee.types import typecompare as bee_typecompare

def typecompare(t1, t2):
  if t1 == "any" or t2 == "any": return True
  return bee_typecompare(t1, t2)

class NodeCanvas(HGui):
  _NodeCanvasClass = None
  def __init__(self, mainWindow, clipboard, statusbar = None):
    assert self._NodeCanvasClass is not None
    self._hNodeCanvas = self._NodeCanvasClass(mainWindow.h(),clipboard,statusbar)   
    mainWindow.setNodeCanvas(self)
    self._hNodeCanvas._hgui = weakref.ref(self)
    
    self._connections = {}
    self._connection_ids = []
    self._nodes = {}
    self.observers_selection = []
    self.observers_remove = []
    self._clipboard = weakref.ref(clipboard)
  
  def add_connection(self, id_, start, end, interpoints = []):
    start_node, start_attribute = start
    end_node, end_attribute = end

    if self._redundant_connection(
     start_node, end_node, 
     start_attribute, end_attribute     
    ):
      return
    
    assert start_node in self._nodes, start_node
    assert end_node in self._nodes, end_node
    
    connection = Connection(
      start_node, start_attribute,
      end_node, end_attribute,
      interpoints
    )

    assert id_ not in self._connections
    self._connections[id_] = connection
    self._connection_ids.append(id_)

    valid = self._valid_connection(connection)
    self._hNodeCanvas.h_add_connection(id_, connection, valid)
  
  def _redundant_connection(self, 
     start_node, end_node, 
     start_attribute, end_attribute 
    ):
    for con in self._connections.values():
      if con.start_node != start_node: continue
      if con.end_node != end_node: continue
      if con.start_attribute != start_attribute: continue
      if con.end_attribute != end_attribute: continue
      return True
    return False
      
  def _valid_connection(self, connection):
    ret = False
    while 1:
      start_node = self._nodes[connection.start_node]    
      start = start_node.get_attribute(connection.start_attribute)
      start = start.outhook
      if start is None: break

      end_node = self._nodes[connection.end_node]
      end = end_node.get_attribute(connection.end_attribute)
      end = end.inhook
      if end is None: break
          
      #unequal modes/types
      if start.mode != end.mode: break
      if not typecompare(start.type, end.type): break
      
      #more than 1 connection to pull antenna
      if end.mode == "pull":
        first_con = True
        for con in self._connections.values():
          if con is connection: continue
          if con.end_node != connection.end_node: continue
          if con.end_attribute != connection.end_attribute: continue
          first_con = False
          break
        if not first_con: 
          break
      
      ret = True
      break
    return ret

  def gui_asks_connection(self, connection):
    start_node, end_node = connection.start_node, connection.end_node
    start_attribute, end_attribute = connection.start_attribute, \
     connection.end_attribute
    
    if self._redundant_connection(
     start_node, end_node, 
     start_attribute, end_attribute     
    ): 
      return False
    
    return self._valid_connection(connection)

  def gui_adds_connection(self, connection, id_, force=False):
    ret = self.gui_asks_connection(connection)
    if ret or force:
      assert id_ not in self._connections
      self._connections[id_] = connection
      self._connection_ids.append(id_)
    return ret

  def gui_offsets_connection_interpoints(self, con_id, offset):
    con = self._connections[con_id]
    dx,dy = offset
    con.interpoints = [(x+dx,y+dy) for x,y in con.interpoints]

  def gui_rearranges_connection(self, con_id, insertionpoint, pos):
    assert pos in ("before", "after"), pos
    old_index = self._connection_ids.index(con_id)
    new_index = self._connection_ids.index(insertionpoint)
    if old_index < new_index: new_index -= 1
    if pos == "after": new_index += 1
    if old_index == new_index: return True
    self._connection_ids.pop(old_index) #removes con_id
    if new_index == len(self._connections):
      self._connection_ids.append(con_id)
    else:
      self._connection_ids.insert(new_index, con_id)
    return True
    
  def remove_connection(self, id_, pass_downward=True):
    self._connections.pop(id_)
    self._connection_ids.remove(id_)
    if pass_downward:
      self._hNodeCanvas.remove_connection(id_)
    
  def gui_removes_connection(self, id_):
    self._connections.pop(id_)
    self._connection_ids.remove(id_)
    return True

  def add_node(self, id_, node):
    assert isinstance(node, Node)
    assert id_ not in self._nodes, id_
    self._nodes[id_] = node
    self._hNodeCanvas.h_add_node(id_, node)
  
  def get_node(self, id_):
    return self._nodes[id_]
  
  def rename_node(self, old_id, new_id, new_name):
    assert old_id in self._nodes, old_id
    assert new_id not in self._nodes, new_id
    node = self._nodes.pop(old_id)
    node.name = new_name
    self._nodes[new_id] = node
    for con in self._connections.values():
      if con.start_node == old_id: con.start_node = new_id
      if con.end_node == old_id: con.end_node = new_id
    self._hNodeCanvas.rename_node(old_id, new_id, new_name)
    self.select([new_id])
  
  def morph_node(self, id_, newnode, maps_in, maps_out):
    assert id_ in self._nodes, id_
    mapcon = []
    for conid in self._connection_ids:
      con = self._connections[conid] 
      if con.start_node == id_:
        a = con.start_attribute
        if maps_out is not None:
          aa = maps_out[a]
          if aa is None: raise KeyError
        else:
          aa = a  
        con.start_attribute = aa
        mapcon.append((a, aa, "out"))
      if con.end_node == id_:
        a = con.end_attribute
        if maps_in is not None:
          aa = maps_in[a]         
          if aa is None: raise KeyError
        else:
          aa = a            
        con.end_attribute = aa        
        mapcon.append((a, aa, "in"))
    self._nodes[id_] = newnode
    self._hNodeCanvas.h_morph_node(id_, newnode, mapcon)
  
  def copy_clipboard(self, nodes):
    self._clipboard().nodecanvas_copy_nodes(nodes)

  def paste_clipboard(self):
    ids = self._clipboard().nodecanvas_paste_nodes()
    if ids is not None: self.select(ids)
  
  def select(self, ids):
    for id_ in ids:
      assert id_ in self._nodes
    self._hNodeCanvas.select(ids)
    
  def gui_selects(self, ids):
    for ob in self.observers_selection:
      ob(ids)
    return True

  def gui_deselects(self):
    for ob in self.observers_selection:
      ob(None)
    return True
  
  def gui_adds_node(self, id_, node):
    assert isinstance(node, Node)
    assert id_ not in self._nodes
    self._nodes[id_] = node
    return True

  def _remove_node(self, id_, pass_downward = True):
    assert id_ in self._nodes, id_
    self._nodes.pop(id_)
    for con_id, connection in list(self._connections.items()):
      if connection.start_node == id_ or \
       connection.end_node == id_:
        self.remove_connection(con_id, pass_downward)

  def remove_node(self, id_):
    self._remove_node(id_)
    self._hNodeCanvas.remove_node(id_)

  def gui_removes_nodes(self, ids):
    for id_ in ids:
      self._remove_node(id_)
      for ob in self.observers_remove:
        ob(id_)
    return True

  def gui_moves_node(self, id_, position):
    node = self._nodes[id_]
    node.position = position
    return True

  def get_connections(self):
    return [self._connections[id_] for id_ in self._connection_ids]
  def get_connection_ids(self):
    return self._connection_ids  
  
  def set_attribute_value(self, id_, attribute, value):
    #nodes themselves are not modified!
    self._hNodeCanvas.set_attribute_value(id_, attribute, value)  
    
  def h(self):
    return self._hNodeCanvas
  
